{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "43ac54ab-e86d-40b0-8188-2638d995ed60",
   "metadata": {},
   "source": [
    "# Importation et quelques tests"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "74e3a057-d1c3-42d7-8646-133275825ea2",
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch\n",
    "print(torch.__version__)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d7e4ac76-6b39-4330-b16b-d5cc56895059",
   "metadata": {},
   "outputs": [],
   "source": [
    "import torchvision\n",
    "print(torchvision.__version__)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "6b00ae57-fee1-4479-ab83-f3c15f3df04b",
   "metadata": {},
   "outputs": [],
   "source": [
    "import os, time\n",
    "\n",
    "import numpy as np\n",
    "import matplotlib.pylab as plt\n",
    "\n",
    "import torch\n",
    "import torch.nn as nn\n",
    "import torch.nn.functional as F\n",
    "from torch import optim\n",
    "import torchvision\n",
    "from torchvision.transforms import v2\n",
    "from torchvision import datasets\n",
    "\n",
    "plt.rcParams['figure.figsize'] = (10, 6)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "293a7f42-8af2-4684-93dd-f380d6e9f87f",
   "metadata": {},
   "outputs": [],
   "source": [
    "torch.cuda.is_available()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "d3dddde6-6638-4ba9-b314-141638013a3a",
   "metadata": {},
   "source": [
    "# Exemple de classification MNIST-Fashion\n",
    "\n",
    "Cette partie sert d’exemple complet et fournit quelques fonctions\n",
    "utilitaires pour la suite.\n",
    "\n",
    "De nombreux datasets sont accessibles directement: voir\n",
    "<https://pytorch.org/docs/stable/torchvision/datasets.html>\n",
    "\n",
    "PyTorch utilise une classe DataLoader qui permet de gérer l’accès aux\n",
    "données (en particulier, en gérant efficamenet le parallélisme et les\n",
    "batchs): voir\n",
    "<https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader>\n",
    "\n",
    "De plus, de nombreuses tranformations sont disponbiles pour réaliser\n",
    "l’augmentation des données: voir\n",
    "<https://pytorch.org/docs/stable/torchvision/transforms.html>\n",
    "\n",
    "On charge ici un dataset prédéfini\n",
    "<https://pytorch.org/docs/stable/torchvision/datasets.html#torchvision.datasets.FashionMNIST>\n",
    "\n",
    "Il s’agit de la base d’images décrite ici:\n",
    "<https://github.com/zalandoresearch/fashion-mnist>"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "4cdb3318-8554-4167-a6d7-b12458de4db2",
   "metadata": {},
   "outputs": [],
   "source": [
    "train_loader = torch.utils.data.DataLoader(\n",
    "    torchvision.datasets.FashionMNIST('fashionmnist', train=True, download=True,\n",
    "                   transform=torchvision.transforms.Compose([\n",
    "                       torchvision.transforms.ToTensor(),\n",
    "                       torchvision.transforms.Normalize((0.1307,), (0.3081,))\n",
    "                       ])),\n",
    "    batch_size=64, shuffle=True)\n",
    "test_loader = torch.utils.data.DataLoader(\n",
    "    torchvision.datasets.FashionMNIST('fashionmnist', train=False,\n",
    "                               transform=torchvision.transforms.Compose([\n",
    "                                   torchvision.transforms.ToTensor(),\n",
    "                                   torchvision.transforms.Normalize((0.1307,), (0.3081,))\n",
    "                                   ])),\n",
    "    batch_size=1000, shuffle=True)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e34ff98a-681b-4f7c-8f3e-48fd34a8e3c7",
   "metadata": {},
   "outputs": [],
   "source": [
    "class Net(nn.Module):\n",
    "    def __init__(self):\n",
    "        super(Net, self).__init__()\n",
    "        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)\n",
    "        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)\n",
    "        self.conv2_drop = nn.Dropout2d()\n",
    "        self.fc1 = nn.Linear(320, 50)\n",
    "        self.fc2 = nn.Linear(50, 10)\n",
    "\n",
    "    def forward(self, x):\n",
    "        x = F.relu(F.max_pool2d(self.conv1(x), 2))\n",
    "        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))\n",
    "        x = x.view(-1, 320)\n",
    "        x = F.relu(self.fc1(x))\n",
    "        x = F.dropout(x, training=self.training)\n",
    "        x = self.fc2(x)\n",
    "\n",
    "        return F.log_softmax(x, dim=1)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "aed80ce5-c162-4808-823a-d352269e7a5c",
   "metadata": {},
   "outputs": [],
   "source": [
    "model = Net()\n",
    "\n",
    "device = torch.device(\"cuda:0\" if torch.cuda.is_available() else \"cpu\")\n",
    "model = model.to(device)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "440d4757-5a99-499a-b193-a61530486855",
   "metadata": {},
   "source": [
    "Optimizer :\n",
    "<https://pytorch.org/docs/stable/optim.html#module-torch.optim>"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "aa5ddda0-4bf0-494f-807f-a5249addb565",
   "metadata": {},
   "outputs": [],
   "source": [
    "optimizer = optim.SGD(model.parameters(), lr=0.01)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "de8f3a94-6d17-4b36-b98d-b39b113e47f7",
   "metadata": {},
   "source": [
    "Fonction de loss\n",
    "<https://pytorch.org/docs/stable/nn.html#loss-functions>"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "325e3e1e-5442-4942-b06b-04b951a9b264",
   "metadata": {},
   "outputs": [],
   "source": [
    "criterion = torch.nn.NLLLoss()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "8c36ab71-8f18-4cf9-b3cf-b9f6d9522415",
   "metadata": {},
   "source": [
    "Fonction de train:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b922a1e4-fc02-420a-b2ca-50ba109e1850",
   "metadata": {},
   "outputs": [],
   "source": [
    "def train(model, device, train_loader, optimizer, epoch, criterion):\n",
    "    model.train()\n",
    "    for batch_idx, (data, target) in enumerate(train_loader):\n",
    "        data, target = data.to(device), target.to(device)\n",
    "        optimizer.zero_grad()\n",
    "\n",
    "        output = model(data)\n",
    "        loss = criterion(output, target)\n",
    "        loss.backward()\n",
    "        optimizer.step()\n",
    "        \n",
    "        if batch_idx % 100 == 0:\n",
    "            print('Train Epoch: {} [{}/{} ({:.0f}%)]\\tLoss: {:.6f}'.format(\n",
    "                epoch, batch_idx * len(data), len(train_loader.dataset),\n",
    "                100. * batch_idx / len(train_loader), loss.item()))"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "794dc87b-9ef4-4aab-b5cc-66b835b1a441",
   "metadata": {},
   "source": [
    "Fonction de test:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a12f2d28-5499-4c89-b926-84c0c591a2c9",
   "metadata": {},
   "outputs": [],
   "source": [
    "def test(model, device, test_loader, criterion):\n",
    "    model.eval()\n",
    "    test_loss = 0\n",
    "    correct = 0\n",
    "    with torch.no_grad():\n",
    "        for data, target in test_loader:\n",
    "            data, target = data.to(device), target.to(device)\n",
    "            output = model(data)\n",
    "            test_loss += criterion(output, target).item() # sum up batch loss\n",
    "            pred = output.max(1, keepdim=True)[1] # get the index of the max log-probability\n",
    "            correct += pred.eq(target.view_as(pred)).sum().item()\n",
    "\n",
    "    test_loss /= len(test_loader.dataset)\n",
    "    print('\\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\\n'.format(\n",
    "        test_loss, correct, len(test_loader.dataset),\n",
    "        100. * correct / len(test_loader.dataset)))\n",
    "    return "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0cb98751-1935-409f-a9b0-f6e06e22fa58",
   "metadata": {},
   "outputs": [],
   "source": [
    "for epoch in range(1, 10):\n",
    "        start_time = time.process_time()\n",
    "        train(model, device, train_loader, optimizer, epoch,criterion)\n",
    "        end_time = time.process_time()\n",
    "        print(\"Train Epoch duration {:.2f}s\".format(end_time - start_time))\n",
    "        \n",
    "        test(model, device, test_loader,criterion)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "df3fcf81-140c-492e-b6c3-63365ba86ad4",
   "metadata": {},
   "source": [
    "    Train Epoch: 1 [0/60000 (0%)]   Loss: 2.331967\n",
    "    Train Epoch: 1 [6400/60000 (11%)]   Loss: 2.220635\n",
    "    Train Epoch: 1 [12800/60000 (21%)]  Loss: 2.072088\n",
    "    Train Epoch: 1 [19200/60000 (32%)]  Loss: 1.749465\n",
    "    Train Epoch: 1 [25600/60000 (43%)]  Loss: 1.258822\n",
    "    Train Epoch: 1 [32000/60000 (53%)]  Loss: 1.271635\n",
    "    Train Epoch: 1 [38400/60000 (64%)]  Loss: 1.171696\n",
    "    Train Epoch: 1 [44800/60000 (75%)]  Loss: 1.161705\n",
    "    Train Epoch: 1 [51200/60000 (85%)]  Loss: 1.034610\n",
    "    Train Epoch: 1 [57600/60000 (96%)]  Loss: 0.904365\n",
    "    Train Epoch duration 95.98s\n",
    "\n",
    "    Test set: Average loss: 0.0008, Accuracy: 6984/10000 (70%)\n",
    "\n",
    "    [...]\n",
    "\n",
    "    Train Epoch: 9 [0/60000 (0%)]   Loss: 0.813664\n",
    "    Train Epoch: 9 [6400/60000 (11%)]   Loss: 0.643290\n",
    "    Train Epoch: 9 [12800/60000 (21%)]  Loss: 0.792572\n",
    "    Train Epoch: 9 [19200/60000 (32%)]  Loss: 0.680120\n",
    "    Train Epoch: 9 [25600/60000 (43%)]  Loss: 0.951693\n",
    "    Train Epoch: 9 [32000/60000 (53%)]  Loss: 0.531021\n",
    "    Train Epoch: 9 [38400/60000 (64%)]  Loss: 0.510627\n",
    "    Train Epoch: 9 [44800/60000 (75%)]  Loss: 0.764454\n",
    "    Train Epoch: 9 [51200/60000 (85%)]  Loss: 0.707269\n",
    "    Train Epoch: 9 [57600/60000 (96%)]  Loss: 0.570231\n",
    "    Train Epoch duration 132.90s\n",
    "\n",
    "    Test set: Average loss: 0.0005, Accuracy: 8193/10000 (82%)\n",
    "\n",
    "## À ajouter\n",
    "\n",
    "Plot en fonction des epoch:\n",
    "\n",
    "-   training loss\n",
    "-   testing loss\n",
    "-   accuracy"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "689f89c2-0dc2-4a34-904b-173f3402b925",
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "markdown",
   "id": "a3db8c06-9589-4794-b643-7a66f1e2e41e",
   "metadata": {},
   "source": [
    "# Fine-tuning\n",
    "\n",
    "Les données sont disponibles à l’adresse\n",
    "<https://download.pytorch.org/tutorial/hymenoptera_data.zip>\n",
    "\n",
    "## Téléchargement des données"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0221748b-e2ef-4b6c-8e7c-3b8ff75a974b",
   "metadata": {},
   "outputs": [],
   "source": [
    "!wget https://download.pytorch.org/tutorial/hymenoptera_data.zip\n",
    "!unzip hymenoptera_data.zip"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "0e3ac9bc-e0f9-4e52-a1dc-e8e56ab09df5",
   "metadata": {},
   "source": [
    "### Augmentation de données basique et très classique"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "339d42bb-650e-4399-a422-5fbda40bf399",
   "metadata": {},
   "outputs": [],
   "source": [
    "data_transforms = {\n",
    "    # Data augmentation and normalization for training\n",
    "    'train': v2.Compose([\n",
    "        v2.RandomResizedCrop(224),\n",
    "        v2.RandomHorizontalFlip(),\n",
    "        v2.ToImage(),\n",
    "        v2.ToDtype(torch.float32, scale=True),\n",
    "        v2.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])\n",
    "    ]),\n",
    "    # Just normalization for validation\n",
    "    'val': v2.Compose([\n",
    "        v2.Resize(256),\n",
    "        v2.CenterCrop(224),\n",
    "        v2.ToImage(),\n",
    "        v2.ToDtype(torch.float32, scale=True),\n",
    "        v2.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])\n",
    "    ]),\n",
    "}"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "34ec5ff4-6f82-41ed-bdf0-f120a68baa87",
   "metadata": {},
   "source": [
    "### Construction d’un dataloader\n",
    "\n",
    "Voir <https://pytorch.org/vision/stable/datasets.html> pour les datasets\n",
    "classiques et aussi\n",
    "<https://pytorch.org/vision/stable/generated/torchvision.datasets.DatasetFolder.html#torchvision.datasets.DatasetFolder>\n",
    "pour faire des datasets personnalisés."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "700dc137-b6e5-4574-b385-4ae75e98dd32",
   "metadata": {},
   "outputs": [],
   "source": [
    "data_dir = 'hymenoptera_data'\n",
    "image_datasets = {x: datasets.ImageFolder(os.path.join(data_dir, x),\n",
    "                                          data_transforms[x])\n",
    "                  for x in ['train', 'val']}\n",
    "dataloaders = {x: torch.utils.data.DataLoader(image_datasets[x], batch_size=4,\n",
    "                                             shuffle=True, num_workers=4)\n",
    "              for x in ['train', 'val']}\n",
    "dataset_sizes = {x: len(image_datasets[x]) for x in ['train', 'val']}\n",
    "class_names = image_datasets['train'].classes\n",
    "\n",
    "device = torch.device(\"cuda:0\" if torch.cuda.is_available() else \"cpu\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "4725e079-06ac-4410-9126-6abfedf8b47d",
   "metadata": {},
   "outputs": [],
   "source": [
    "def imshow(inp, title=None):\n",
    "    \"\"\"Imshow for Tensor.\"\"\"\n",
    "    inp = inp.numpy().transpose((1, 2, 0))\n",
    "    mean = np.array([0.485, 0.456, 0.406])\n",
    "    std = np.array([0.229, 0.224, 0.225])\n",
    "    inp = std * inp + mean\n",
    "    inp = np.clip(inp, 0, 1)\n",
    "    plt.imshow(inp)\n",
    "    if title is not None:\n",
    "        plt.title(title)\n",
    "    plt.pause(0.001)  # pause a bit so that plots are updated\n",
    "\n",
    "\n",
    "# Get a batch of training data\n",
    "inputs, classes = next(iter(dataloaders['train']))\n",
    "\n",
    "# Make a grid from batch\n",
    "out = torchvision.utils.make_grid(inputs)\n",
    "\n",
    "imshow(out, title=[class_names[x] for x in classes])"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e9dca0ae-785a-4bb8-823f-b519cbd81794",
   "metadata": {},
   "source": [
    "## Resnet\n",
    "\n",
    "Voir\n",
    "<https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py>\n",
    "pour comprendre comment couper le réseau."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "6cf27fb8-d4d8-4796-8139-c82c5fe3fe86",
   "metadata": {},
   "outputs": [],
   "source": [
    "from torchvision.models import resnet18\n",
    "\n",
    "model = resnet18(weights='IMAGENET1K_V1') # Modèle pré-entraîné sur ImageNet"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b3492db2-57b2-4a74-ba4a-04c46a414410",
   "metadata": {},
   "outputs": [],
   "source": [
    "print(model)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "bd3edc39-07e8-4589-b08d-4f4ce89846bb",
   "metadata": {},
   "source": [
    "On fixe les paramètres de toutes les couches, ils ne seront pas modifiés\n",
    "lors de l’apprentissage."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9526a7fe-7866-42c4-965f-929d2b0f73b9",
   "metadata": {},
   "outputs": [],
   "source": [
    "for param in model.parameters():\n",
    "    param.requires_grad = False"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "0f7790a4-8193-4232-be1c-b8d5ead3d623",
   "metadata": {},
   "source": [
    "On remplace la partie finale par une simple couche dense (qui sera elle\n",
    "apprise)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "03515bb9-1a61-4010-bc14-6e448cd7d3be",
   "metadata": {},
   "outputs": [],
   "source": [
    "model.fc = nn.Linear(512, 2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e2e675eb-5fd6-45e9-8162-a5d544e24a70",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Calcul de la représentation, pour l'envoyer à un SVM par exemple\n",
    "# model.avgpool(x) # sortie de dernièer couche avant le classifier\n",
    "# Ce serait un prétraitement à appliquer à toutes les données avant d'entraîner le SVM"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "bff7ca43-e4f6-40aa-8a3a-ffc38eff169a",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "ResNet(\n",
       "  (conv1): Conv2d(3, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False)\n",
       "  (bn1): BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "  (relu): ReLU(inplace=True)\n",
       "  (maxpool): MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False)\n",
       "  (layer1): Sequential(\n",
       "    (0): BasicBlock(\n",
       "      (conv1): Conv2d(64, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn1): BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      (relu): ReLU(inplace=True)\n",
       "      (conv2): Conv2d(64, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn2): BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "    )\n",
       "    (1): BasicBlock(\n",
       "      (conv1): Conv2d(64, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn1): BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      (relu): ReLU(inplace=True)\n",
       "      (conv2): Conv2d(64, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn2): BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "    )\n",
       "  )\n",
       "  (layer2): Sequential(\n",
       "    (0): BasicBlock(\n",
       "      (conv1): Conv2d(64, 128, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False)\n",
       "      (bn1): BatchNorm2d(128, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      (relu): ReLU(inplace=True)\n",
       "      (conv2): Conv2d(128, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn2): BatchNorm2d(128, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      (downsample): Sequential(\n",
       "        (0): Conv2d(64, 128, kernel_size=(1, 1), stride=(2, 2), bias=False)\n",
       "        (1): BatchNorm2d(128, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      )\n",
       "    )\n",
       "    (1): BasicBlock(\n",
       "      (conv1): Conv2d(128, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn1): BatchNorm2d(128, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      (relu): ReLU(inplace=True)\n",
       "      (conv2): Conv2d(128, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn2): BatchNorm2d(128, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "    )\n",
       "  )\n",
       "  (layer3): Sequential(\n",
       "    (0): BasicBlock(\n",
       "      (conv1): Conv2d(128, 256, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False)\n",
       "      (bn1): BatchNorm2d(256, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      (relu): ReLU(inplace=True)\n",
       "      (conv2): Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn2): BatchNorm2d(256, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      (downsample): Sequential(\n",
       "        (0): Conv2d(128, 256, kernel_size=(1, 1), stride=(2, 2), bias=False)\n",
       "        (1): BatchNorm2d(256, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      )\n",
       "    )\n",
       "    (1): BasicBlock(\n",
       "      (conv1): Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn1): BatchNorm2d(256, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      (relu): ReLU(inplace=True)\n",
       "      (conv2): Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn2): BatchNorm2d(256, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "    )\n",
       "  )\n",
       "  (layer4): Sequential(\n",
       "    (0): BasicBlock(\n",
       "      (conv1): Conv2d(256, 512, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False)\n",
       "      (bn1): BatchNorm2d(512, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      (relu): ReLU(inplace=True)\n",
       "      (conv2): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn2): BatchNorm2d(512, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      (downsample): Sequential(\n",
       "        (0): Conv2d(256, 512, kernel_size=(1, 1), stride=(2, 2), bias=False)\n",
       "        (1): BatchNorm2d(512, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      )\n",
       "    )\n",
       "    (1): BasicBlock(\n",
       "      (conv1): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn1): BatchNorm2d(512, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "      (relu): ReLU(inplace=True)\n",
       "      (conv2): Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)\n",
       "      (bn2): BatchNorm2d(512, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)\n",
       "    )\n",
       "  )\n",
       "  (avgpool): AdaptiveAvgPool2d(output_size=(1, 1))\n",
       "  (fc): Linear(in_features=512, out_features=1, bias=True)\n",
       ")"
      ]
     },
     "execution_count": 0,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "optimizer = optim.SGD(model.fc.parameters(), lr=0.01)\n",
    "\n",
    "criterion = nn.CrossEntropyLoss()\n",
    "device = torch.device(\"cuda:0\" if torch.cuda.is_available() else \"cpu\")\n",
    "model.to(device)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "40693e4d-a385-4903-94cd-baeda6070bd7",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Train Epoch: 1 [0/244 (0%)] Loss: 9.530968\n",
      "Train Epoch duration 150.19s\n",
      "\n",
      "Test set: Average loss: 2.7408, Accuracy: 0/153 (0%)\n",
      "\n",
      "Train Epoch: 2 [0/244 (0%)] Loss: 9.099474"
     ]
    }
   ],
   "source": [
    "for epoch in range(1, 10):\n",
    "        start_time = time.process_time()\n",
    "        train(model, device, dataloaders[\"train\"], optimizer, epoch, criterion)\n",
    "        end_time = time.process_time()\n",
    "        print(\"Train Epoch duration {:.2f}s\".format(end_time - start_time))\n",
    "        \n",
    "        test(model, device, dataloaders[\"val\"], criterion)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "8011b610-5168-4784-b399-00e985b757d8",
   "metadata": {},
   "source": [
    "## VGG\n",
    "\n",
    "Voir\n",
    "<https://github.com/pytorch/vision/blob/master/torchvision/models/vgg.py>\n",
    "et\n",
    "<https://pytorch.org/docs/stable/_modules/torch/nn/modules/container.html#Sequential>"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "8c117da4-b0f8-4120-ab66-f0234350ff53",
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.7"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
